"""
Módulo que realiza o parser de um inventário remoto (no SCVMM)
"""
import json
from vmm_manager.infra.comando import Comando
from vmm_manager.util.config import (CAMPO_AGRUPAMENTO, CAMPO_ID, CAMPO_IMAGEM,
                                     CAMPO_REGIAO, CAMPO_REDE_PRINCIPAL)
from vmm_manager.entidade.inventario import Inventario
from vmm_manager.entidade.vm import VM
from vmm_manager.entidade.vm_rede import VMRede
from vmm_manager.scvmm.enums import VMStatusEnum


class ParserRemoto:
    def __init__(self, agrupamento, nuvem):
        self.agrupamento = agrupamento
        self.nuvem = nuvem
        self.__inventario = None

    def __get_vms_servidor(self, servidor_acesso, filtro_nome_vm=None):
        cmd = Comando('obter_vms_agrupamento',
                      servidor_vmm=servidor_acesso.servidor_vmm,
                      campo_agrupamento=CAMPO_AGRUPAMENTO[0],
                      campo_id=CAMPO_ID[0],
                      campo_imagem=CAMPO_IMAGEM[0],
                      campo_regiao=CAMPO_REGIAO[0],
                      campo_rede_principal=CAMPO_REDE_PRINCIPAL[0],
                      agrupamento=self.agrupamento,
                      filtro_nome_vm=filtro_nome_vm,
                      nuvem=self.nuvem)
        status, vms = cmd.executar(servidor_acesso)
        if not status:
            raise Exception(
                "Erro ao recuperar VM's do agrupamento: {}".format(vms))
        return vms

    def __montar_inventario(self, servidor_acesso, filtro_nome_vm=None):
        self.__inventario = Inventario(self.agrupamento, self.nuvem)
        vms_servidor = json.loads(
            self.__get_vms_servidor(servidor_acesso, filtro_nome_vm) or '{}')
        for maquina_virtual in vms_servidor:
            vms_rede = []
            for rede in maquina_virtual.get('Redes'):
                vm_rede = VMRede(rede.get('Nome'), rede.get('Principal'))
                vm_rede.ips = rede.get('IPS', '').split(' ')
                vms_rede.append(vm_rede)

            self.__inventario.vms[maquina_virtual.get('Nome')] = VM(
                maquina_virtual.get('Nome'),
                maquina_virtual.get('Descricao'),
                maquina_virtual.get('Imagem'),
                maquina_virtual.get('Regiao'),
                maquina_virtual.get('QtdeCpu'),
                maquina_virtual.get('QtdeRam'),
                vms_rede,
                maquina_virtual.get('ID'),
                VMStatusEnum(maquina_virtual.get('Status')),
                maquina_virtual.get('NoRegiao'),
            )

    def get_inventario(self, servidor_acesso, filtro_nome_vm=None):
        if not self.__inventario:
            try:
                self.__montar_inventario(servidor_acesso, filtro_nome_vm)
            # pylint: disable=broad-except
            except Exception as ex:
                return False, str(ex)

        return True, self.__inventario



"""
Representação de um inventário
"""
import json
from vmm_manager.entidade.plano_execucao import PlanoExecucao
from vmm_manager.entidade.acao import Acao
from vmm_manager.util.config import CAMPO_AGRUPAMENTO
from vmm_manager.infra.comando import Comando


def json_handle_inventario(obj):
    if isinstance(obj, Inventario):
        return obj.to_dict()
    raise ValueError('Objeto precisa ser uma instância de inventário.')


class Inventario:
    REGIAO_PADRAO = 'default'

    @staticmethod
    def get_json(inventario_local, inventario_remoto):
        for nome_vm in inventario_remoto.vms:
            if nome_vm not in inventario_local.vms:
                # máquina órfã: não exibir
                continue

            inventario_remoto.vms[nome_vm].dados_ansible = \
                inventario_local.vms[nome_vm].dados_ansible

        return True, json.dumps(inventario_remoto,
                                default=json_handle_inventario,
                                sort_keys=True, indent=4)

    def __init__(self, agrupamento, nuvem):
        self.agrupamento = agrupamento
        self.nuvem = nuvem
        self.vms = {}

    def calcular_plano_execucao(self, inventario_remoto):
        if (self.agrupamento != inventario_remoto.agrupamento
                or self.nuvem != inventario_remoto.nuvem):
            return False, 'Não é possível calcular o plano de execução \
                para inventários de agrupamento ou nuvem distintos.'

        plano_execucao = PlanoExecucao(self.agrupamento, self.nuvem)

        self.__add_acoes_criar_vms(inventario_remoto, plano_execucao)
        self.__add_acoes_execucao_excluir_vms(
            inventario_remoto, plano_execucao)

        return True, plano_execucao

    def gerar_plano_exclusao(self):
        plano_execucao = PlanoExecucao(self.agrupamento, self.nuvem)

        for maquina_virtual in self.vms.values():
            plano_execucao.acoes.append(
                Acao('excluir_vm',
                     id_vmm=maquina_virtual.id_vmm
                     )
            )

        return plano_execucao

    def is_vazio(self):
        return not self.vms

    def lista_nome_vms_str(self):
        return ','.join(['"{}"'.format(nome_vm) for nome_vm in self.vms])

    def validar_no_servidor(self, servidor_acesso):
        imagens = set()
        redes = set()
        regioes = set()

        for maquina_virtual in self.vms.values():
            if maquina_virtual.imagem is None:
                raise ValueError(
                    'Imagem da VM {} não definida.'.format(maquina_virtual.nome))
            imagens.add(maquina_virtual.imagem)

            if maquina_virtual.regiao != Inventario.REGIAO_PADRAO:
                regioes.add(maquina_virtual.regiao)

            if maquina_virtual.qtde_cpu is None:
                raise ValueError(
                    'Quantidade de CPUs da VM {} não definida.'.format(maquina_virtual.nome))

            if maquina_virtual.qtde_ram_mb is None:
                raise ValueError(
                    'Quantidade de memória da VM {} não definida.'.format(maquina_virtual.nome))

            if maquina_virtual.get_qtde_rede_principal() != 1:
                raise ValueError(
                    'VM {} deve ter exatamente uma rede principal.'.format(maquina_virtual.nome))

            redes.update([rede.nome for rede in maquina_virtual.redes])

        cmd = Comando('validar_inventario', imagens=imagens,
                      nuvem=self.nuvem,
                      redes=redes,
                      servidor_vmm=servidor_acesso.servidor_vmm,
                      qtde_minima_regioes=len(regioes),
                      agrupamento=self.agrupamento,
                      lista_nome_vms_str=self.lista_nome_vms_str(),
                      campo_agrupamento=CAMPO_AGRUPAMENTO[0])
        _, msg = cmd.executar(servidor_acesso)
        if msg:
            raise ValueError(msg)

    def __add_acoes_criar_vms(self, inventario_remoto, plano_execucao):
        vms_inserir = [
            nome_vm_local for nome_vm_local in self.vms
            if nome_vm_local not in inventario_remoto.vms
        ]
        for nome_vm in vms_inserir:
            plano_execucao.acoes.append(
                Acao('criar_vm',
                     nome=self.vms[nome_vm].nome,
                     descricao=self.vms[nome_vm].descricao,
                     imagem=self.vms[nome_vm].imagem,
                     regiao=self.vms[nome_vm].regiao,
                     qtde_cpu=self.vms[nome_vm].qtde_cpu,
                     qtde_ram_mb=self.vms[nome_vm].qtde_ram_mb,
                     redes=[rede.nome for rede in self.vms[nome_vm].redes],
                     rede_principal=self.vms[nome_vm].get_rede_principal()
                     )
            )

    def __add_acoes_execucao_excluir_vms(self, inventario_remoto, plano_execucao):
        vms_excluir = [
            nome_vm_remoto for nome_vm_remoto in inventario_remoto.vms
            if nome_vm_remoto not in self.vms
        ]
        for nome_vm in vms_excluir:
            plano_execucao.acoes.append(
                Acao('excluir_vm',
                     id_vmm=inventario_remoto.vms[nome_vm].id_vmm
                     )
            )

    def __eq__(self, other):
        return isinstance(other, Inventario) and (self.agrupamento == other.agrupamento
                                                  and self.nuvem == other.nuvem
                                                  and self.vms == other.vms)

    def __str__(self):
        return '''
            agrupamento: {}
            nuvem: {}
            vms: {}
            '''.format(self.agrupamento,
                       self.nuvem,
                       self.vms)

    def to_dict(self):
        return {
            'agrupamento': self.agrupamento,
            'nuvem': self.nuvem,
            'vms': [self.vms[vm_name].to_dict() for vm_name in self.vms]
        }



"""
Representação de uma máquina virtual
"""
from vmm_manager.scvmm.enums import VMStatusEnum
from vmm_manager.entidade.vm_ansible import VMAnsible


class VM:
    def __init__(self, nome, descricao,
                 imagem, regiao,
                 qtde_cpu, qtde_ram_mb, redes,
                 id_vmm=None,
                 status=VMStatusEnum.EM_EXECUCAO,
                 no_regiao=None):
        self.nome = nome
        self.descricao = descricao
        self.imagem = imagem
        self.regiao = regiao
        self.qtde_cpu = qtde_cpu
        self.qtde_ram_mb = qtde_ram_mb
        self.redes = redes
        self.id_vmm = id_vmm
        self.status = status
        self.no_regiao = no_regiao
        self.dados_ansible = {}

    def extrair_dados_ansible_dict(self, dict_ansible):
        for item in dict_ansible or {}:
            grupo = item.get('grupo')

            if grupo in self.dados_ansible:
                raise ValueError(
                    "Grupo ansible '{}' referenciado mais de uma vez para a VM '{}'.".format(
                        grupo, self.nome))

            ansible_grupo = VMAnsible(grupo)
            ansible_grupo.extrair_dados_vars_dict(
                item.get('vars'), self.nome)

            self.dados_ansible[grupo] = ansible_grupo

    def get_qtde_rede_principal(self):
        return sum([1 for rede in self.redes if rede.principal])

    def get_rede_principal(self):
        return next((rede.nome for rede in self.redes if rede.principal), None)

    def __hash__(self):
        return hash(self.nome)

    def __eq__(self, other):
        return isinstance(other, VM) and (self.nome == other.nome
                                          and self.imagem == other.imagem
                                          and self.regiao == other.regiao
                                          and self.qtde_cpu == other.qtde_cpu
                                          and self.qtde_ram_mb == other.qtde_ram_mb
                                          and self.redes == other.redes
                                          and self.status == other.status)

    def __repr__(self):
        return '''
                nome: {}
                descricao: {}
                imagem: {}
                regiao: {}
                qtde_cpu: {}
                qtde_ram_mb: {}
                redes: {}
                id_vmm: {}
                status: {}
                no_regiao: {}
                ansible: {}
                '''.format(self.nome,
                           self.descricao,
                           self.imagem,
                           self.regiao,
                           self.qtde_cpu,
                           self.qtde_ram_mb,
                           self.redes,
                           self.id_vmm,
                           self.status,
                           self.no_regiao,
                           self.dados_ansible)

    def to_dict(self):
        return {
            'nome': self.nome,
            'descricao': self.descricao,
            'imagem': self.imagem,
            'regiao': self.regiao,
            'qtde_cpu': self.qtde_cpu,
            'qtde_ram_mb': self.qtde_ram_mb,
            'redes': [rede.to_dict() for rede in self.redes],
            'id_vmm': self.id_vmm,
            'status': self.status.value,
            'no_regiao': self.no_regiao,
            'ansible': [self.dados_ansible[dados_ansible].to_dict()
                        for dados_ansible in self.dados_ansible]
        }



"""
Enums do VMM
"""
from enum import Enum


class VMStatusEnum(Enum):
    EM_EXECUCAO = 0
    DESLIGADA = 1


class SCJobStatusEnum(Enum):
    EM_EXECUCAO = 1
    FALHA = 3
    CANCELADO = 4
    SUCESSO = 5
    SUCESSO_COM_AVISO = 6



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vmm_manager',
 'vmm_manager.entidade',
 'vmm_manager.infra',
 'vmm_manager.parser',
 'vmm_manager.scvmm',
 'vmm_manager.util']

package_data = \
{'': ['*'], 'vmm_manager': ['includes/*', 'includes/ps_templates/*']}

install_requires = \
['configargparse>=1.2,<2.0',
 'jinja2>=2.11,<3.0',
 'paramiko>=2.7,<3.0',
 'pytz>=2020,<2021',
 'ruamel.yaml>=0.16,<0.17',
 'tqdm>=4.46,<5.0',
 'yamale>=2.0,<3.0',
 'yamlable>=1.0,<2.0']

entry_points = \
{'console_scripts': ['vmm_manager = vmm_manager.vmm_manager:main']}

setup_kwargs = {
    'name': 'vmm-manager',
    'version': '0.1.0b3',
    'description': 'Management of resources on System Center Virtual Machine Manager (SCVMM) in a declarative way.',
    'long_description': '# vmm-manager\n\nScript python que gerencia recursos no System Center Virtual Machine Manager (SCVMM), de forma declarativa, com base em um arquivo de configuração YAML.\n\n[![License](https://img.shields.io/github/license/MP-ES/vmm_manager.svg)](LICENSE)\n[![Tests](https://github.com/MP-ES/vmm_manager/workflows/Tests/badge.svg)](https://github.com/MP-ES/vmm_manager/actions?query=workflow%3ATests)\n[![Release](https://github.com/MP-ES/vmm_manager/workflows/Release/badge.svg)](https://github.com/MP-ES/vmm_manager/actions?query=workflow%3ARelease)\n[![Python](https://img.shields.io/pypi/pyversions/vmm-manager.svg)](https://pypi.python.org/pypi/vmm-manager)\n[![PyPI](http://img.shields.io/pypi/v/vmm-manager.svg)](https://pypi.python.org/pypi/vmm-manager)\n\n## Pré-requisitos\n\nÉ necessário ter uma máquina Windows, que servirá como ponto de acesso ao SCVMM, com as seguintes ferramentas:\n\n- OpenSSH\n- Módulo PowerShell do SCVMM (**virtualmachinemanager**), geralmente instalado junto com o Console do Virtual Machine Manager (VMM)\n  \nNessa máquina, também é necessário executar o comando PowerShell `set-executionpolicy unrestricted`, com poderes administrativos.\n\n## Instalação\n\n```shell\npip install -U vmm-manager\n```\n\n## Uso\n\nPara consultar as funções e os parâmetros disponíveis, utilize o comando:\n\n```shell\nvmm_manager -h\n```\n\n### Exemplo de arquivo de inventário\n\n```yaml\nagrupamento: vmm_manager_test\nnuvem: "developer"\nimagem_padrao: "vm_linux"\nqtde_cpu_padrao: 1\nqtde_ram_mb_padrao: 512\nredes_padrao:\n  - nome: "vlan1"\n    principal: true\nvms:\n  - nome: VMM_TEST1\n    descricao: "Test VM"\n    redes:\n      - nome: "vlan1"\n        principal: true\n      - nome: "vlan2"\n    regiao: A\n  - nome: VMM_TEST2\n    regiao: B\n    ansible:\n      - grupo: \'web_server\'\n  - nome: VMM_TEST3\n    ansible:\n      - grupo: \'database\'\n        vars:\n          - nome: \'data_dir\'\n            valor: \'/mnt/data\'\n      - grupo: \'load_balancer\'\n```\n\n## Desenvolvimento\n\n### Instalação e configuração do python-poetry\n\nExecute os comandos a seguir:\n\n```shell\n# instalar o poetry\ncurl -sSL https://raw.githubusercontent.com/sdispater/poetry/master/get-poetry.py | python\necho \'source $HOME/.poetry/env\' >>~/.bashrc\n\n# Configurar autocomplete\n# Bash\npoetry completions bash | sudo tee /etc/bash_completion.d/poetry.bash-completion\n```\n\n### Variáveis de ambiente\n\nDefina as variáveis de ambiente de acordo com as instruções do arquivo **.env.default**. Você pode criar um arquivo **.env** e executar o comando `export $(cat .env | xargs)` para defini-las antes da execução do script.\n\n### Como executar\n\n```shell\n# Carregando envs (opcional)\nexport $(cat .env | xargs)\n\n# Instalando dependências\npoetry install --no-root\n\n# Executando script\npoetry run python -m vmm_manager -h\n```\n\n### Comandos úteis para DEV\n\n```shell\n# Habilitar shell\npoetry shell\n\n# Incluir uma dependência\npoetry add <pacote> [--dev]\n\n# Executar lint\npylint tests/* vmm_manager/*\n\n# Executar testes\npython -m pytest -v\n```\n\n## Referências\n\n- [Virtual Machine Manager](https://docs.microsoft.com/en-us/powershell/module/virtualmachinemanager/?view=systemcenter-ps-2019)\n- [Poetry](https://python-poetry.org/)\n',
    'author': 'Estevão Costa',
    'author_email': 'ecosta@mpes.mp.br',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MP-ES/vmm_manager',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Representação de uma ação (um item do plano de execução)
"""
import json
from yamlable import yaml_info, YamlAble
from vmm_manager.infra.comando import Comando
from vmm_manager.util.config import (CAMPO_AGRUPAMENTO, CAMPO_ID,
                                     CAMPO_IMAGEM, CAMPO_REGIAO, CAMPO_REDE_PRINCIPAL)


@yaml_info(yaml_tag_ns='scvmm_manager')
class Acao(YamlAble):
    def __init__(self, nome_comando, **kwargs):
        self.nome_comando = nome_comando
        self.args = kwargs

        self.__status_execucao = None
        self.__retorno_execucao = None

    def executar(self, agrupamento, nuvem, servidor_acesso, guid):
        cmd = Comando(self.nome_comando,
                      agrupamento=agrupamento,
                      campo_agrupamento=CAMPO_AGRUPAMENTO[0],
                      nuvem=nuvem,
                      guid=guid,
                      servidor_vmm=servidor_acesso.servidor_vmm)
        cmd.args.update(self.args)

        status, retorno = cmd.executar(servidor_acesso)

        self.__status_execucao = status
        self.__retorno_execucao = json.loads(
            retorno) if self.__status_execucao else retorno

        return self.__status_execucao, self.__retorno_execucao

    def is_criacao_vm(self):
        return self.nome_comando == 'criar_vm'

    def has_cmd_pos_execucao(self):
        return self.is_criacao_vm()

    def was_executada_com_sucesso(self):
        return (self.__status_execucao and
                self.__retorno_execucao.get('Status') == 'OK')

    def get_resultado_execucao(self):
        if self.__status_execucao is None:
            raise AttributeError('Ação não executada.')
        if self.__status_execucao:
            return self.__retorno_execucao
        return {'Msgs': self.__retorno_execucao}

    def get_cmd_pos_execucao(self, agrupamento, servidor_acesso):
        if self.is_criacao_vm():
            cmd = Comando('criar_vm_pos',
                          descricao='tagueamento da VM {}'.format(
                              self.args['nome']),
                          servidor_vmm=servidor_acesso.servidor_vmm,
                          campo_agrupamento=CAMPO_AGRUPAMENTO[0],
                          campo_id=CAMPO_ID[0],
                          campo_imagem=CAMPO_IMAGEM[0],
                          campo_regiao=CAMPO_REGIAO[0],
                          campo_rede_principal=CAMPO_REDE_PRINCIPAL[0],
                          agrupamento=agrupamento)
            cmd.args.update(self.args)
            return cmd

        raise AttributeError(
            'Ação "{}" não possui comando de pós-execução.'.format(self.nome_comando))

    def get_str_impressao_inline(self):
        return '{} - [{}]'.format(self.nome_comando,
                                  ', '.join(
                                      ['{}={}'.format(arg, self.args[arg])
                                       for arg in self.args]))

    def __eq__(self, other):
        return isinstance(other, Acao) and (self.nome_comando == other.nome_comando
                                            and self.args == other.args)

    def __str__(self):
        return '''
            nome_comando: {}
            args: {}
            '''.format(self.nome_comando,
                       self.args)

    def __to_yaml_dict__(self):
        return {'nome_comando': self.nome_comando,
                'args': self.args}



"""
Script que gerencia um inventário de máquinas no SCVMM, com base
em um arquivo YAML.
"""
import re
import string
import os
import argparse
from distutils.util import strtobool
import configargparse
from ruamel.yaml import YAML, scanner
from vmm_manager.infra.servidor_acesso import ServidorAcesso
from vmm_manager.infra.comando import Comando
from vmm_manager.parser.parser_local import ParserLocal
from vmm_manager.parser.parser_remoto import ParserRemoto
from vmm_manager.entidade.plano_execucao import PlanoExecucao
from vmm_manager.util.config import(CAMPO_AGRUPAMENTO, CAMPO_ID, CAMPO_IMAGEM,
                                    CAMPO_REGIAO, CAMPO_REDE_PRINCIPAL)
from vmm_manager.util.msgs import finalizar_com_erro, imprimir_acao_corrente
from vmm_manager.util.operacao import validar_retorno_operacao_com_lock
from vmm_manager.util.operacao import validar_retorno_operacao_sem_lock
from vmm_manager.util.operacao import adquirir_lock, liberar_lock, confirmar_acao_usuario_com_lock
from vmm_manager.entidade.inventario import Inventario


def parametro_arquivo_yaml(nome_arquivo):
    try:
        with open(nome_arquivo, 'r') as stream:
            yaml = YAML()
            yaml.load(stream)
            return nome_arquivo
    except FileNotFoundError as exc:
        raise argparse.ArgumentTypeError(exc)
    except scanner.ScannerError as exc:
        raise argparse.ArgumentTypeError(
            'YAML inválido: {}'.format(exc))


def parametro_booleano(valor):
    try:
        return strtobool(valor)
    except ValueError:
        raise argparse.ArgumentTypeError(
            "'{}' não é um valor booleano".format(valor))


def parametro_alfanumerico_limitado(valor):
    regex_alfa_num = re.compile(r'^[a-z]{1}[a-z0-9_]{2,39}$', re.IGNORECASE)
    if valor and not regex_alfa_num.match(valor):
        raise argparse.ArgumentTypeError(
            "Parâmetro inválido: '{}'".format(valor))
    return valor


def get_parser():
    parser = configargparse.ArgumentParser(
        description='''
        Script python que gerencia recursos no System Center Virtual Machine Manager (SCVMM), \
            de forma declarativa, com base em um arquivo de configuração YAML.
        ''', default_config_files=['vmm_manager.ini'])
    parser.add('-a', '--servidor-acesso',
               help='''
               Endereço (FQDN ou IP) do servidor Windows, com o módulo PSH do SCVMM instalado, \
                   que executará os scripts PowerShell. \
                   Ex: windows_host.domain.com
               ''',
               env_var='VMM_SERVIDOR_ACESSO', required=True, type=str)
    parser.add('-u', '--usuario',
               help='Usuário com permissões no servidor de acesso e no SCVMM',
               env_var='VMM_USUARIO', required=True, type=str)
    parser.add('-p', '--senha',
               help='Senha do usuário',
               env_var='VMM_SENHA', required=True, type=str)
    parser.add('-s', '--servidor',
               help='Servidor SCVMM. Ex: scvmm_server.domain.com',
               env_var='VMM_SERVIDOR', required=True, type=str)
    parser.add('-o', '--ocultar-progresso',
               help='Não imprime informações de progresso do comando',
               action='store_true')

    subprasers = parser.add_subparsers(dest='comando')
    plan = subprasers.add_parser(
        'plan', help='Cria um plano de execução, \
            com ações necessárias à sincronização do inventário')
    plan.add_argument('--inventario',
                      help='Arquivo YAML com a especificação das máquinas',
                      dest='arquivo_inventario', env_var='VMM_INVENTARIO',
                      required=True, type=parametro_arquivo_yaml)

    apply = subprasers.add_parser(
        'apply', help='Aplica as alterações necessárias no SCVMM')
    apply.add_argument('--plano-execucao',
                       help='Arquivo YAML com o plano de execução',
                       dest='arquivo_plano_execucao', env_var='VMM_PLANO_EXECUCAO',
                       required=False, type=parametro_arquivo_yaml)
    apply.add_argument('--pular-confirmacao',
                       help='Se for True, o plano de execução é aplicado sem confirmação',
                       dest='pular_confirmacao', env_var='VMM_PULAR_CONFIRMACAO',
                       required=False, type=parametro_booleano)
    apply.add_argument('--inventario',
                       help='Arquivo YAML com a especificação das máquinas',
                       dest='arquivo_inventario', env_var='VMM_INVENTARIO',
                       required=False, type=parametro_arquivo_yaml)

    destroy = subprasers.add_parser(
        'destroy', help='Remove todas as máquinas do agrupamento informado')
    destroy.add_argument(
        '--agrupamento', help='Agrupamento a ser excluído', required=True)
    destroy.add_argument(
        '--nuvem', help='Nuvem a ser analisada', required=True)
    destroy.add_argument('--pular-confirmacao',
                         help='Se for True, o agrupamento é excluído sem confirmação',
                         dest='pular_confirmacao', env_var='VMM_PULAR_CONFIRMACAO',
                         required=False, type=parametro_booleano)

    subprasers.add_parser(
        'opts', help='Lista as opções disponíveis no SCVMM para cada parâmetro da VM')

    show = subprasers.add_parser(
        'show', help='Imprime json com os dados e situação dos ativos do inventário')
    show.add_argument('--inventario',
                      help='Arquivo YAML com a especificação das máquinas',
                      dest='arquivo_inventario', env_var='VMM_INVENTARIO',
                      required=True, type=parametro_arquivo_yaml)
    show.add_argument('--vm',
                      help='Nome da máquina virtual', default='',
                      dest='nome_vm', required=False,
                      type=parametro_alfanumerico_limitado)

    return parser


def obter_inventario_remoto(servidor_acesso, agrupamento, nuvem,
                            ocultar_progresso, filtro_nome_vm=None):
    imprimir_acao_corrente('Obtendo inventário remoto', ocultar_progresso)

    parser_remoto = ParserRemoto(agrupamento, nuvem)
    status, inventario_remoto = parser_remoto.get_inventario(
        servidor_acesso, filtro_nome_vm)
    validar_retorno_operacao_com_lock(status, inventario_remoto,
                                      servidor_acesso, agrupamento,
                                      nuvem, ocultar_progresso)

    return inventario_remoto


def obter_inventario_local(servidor_acesso, arquivo_inventario,
                           ocultar_progresso, filtro_nome_vm=None):
    imprimir_acao_corrente('Obtendo inventário local', ocultar_progresso)

    parser_local = ParserLocal(arquivo_inventario)
    status, inventario_local = parser_local.get_inventario(
        servidor_acesso, filtro_nome_vm)
    validar_retorno_operacao_sem_lock(
        status, inventario_local, ocultar_progresso)

    return inventario_local


def obter_plano_execucao(servidor_acesso, inventario_local, ocultar_progresso):
    inventario_remoto = obter_inventario_remoto(
        servidor_acesso, inventario_local.agrupamento,
        inventario_local.nuvem, ocultar_progresso)

    imprimir_acao_corrente('Calculando plano de execução', ocultar_progresso)
    status, plano_execucao = inventario_local.calcular_plano_execucao(
        inventario_remoto)
    validar_retorno_operacao_com_lock(status, plano_execucao, servidor_acesso,
                                      inventario_local.agrupamento,
                                      inventario_local.nuvem, ocultar_progresso)
    return plano_execucao


def validar_conexao(servidor_acesso, ocultar_progresso):
    imprimir_acao_corrente('Validando conexão com o VMM', ocultar_progresso)
    cmd = Comando('testar_conexao', servidor_vmm=servidor_acesso.servidor_vmm)
    status, msg = cmd.executar(servidor_acesso)

    if status and not 'True' in msg:
        status = False
    validar_retorno_operacao_sem_lock(status, msg, ocultar_progresso)


def configurar_vmm(servidor_acesso, ocultar_progresso):
    imprimir_acao_corrente('Configurando VMM', ocultar_progresso)
    cmd = Comando('configurar_vmm', servidor_vmm=servidor_acesso.servidor_vmm,
                  campos_customizados=[CAMPO_AGRUPAMENTO, CAMPO_ID,
                                       CAMPO_IMAGEM, CAMPO_REGIAO,
                                       CAMPO_REDE_PRINCIPAL])
    status, msg = cmd.executar(servidor_acesso)

    validar_retorno_operacao_sem_lock(status, msg, ocultar_progresso)


def listar_opcoes(servidor_acesso, ocultar_progresso):
    imprimir_acao_corrente('Listando opções', ocultar_progresso)
    cmd = Comando('listar_opcoes', servidor_vmm=servidor_acesso.servidor_vmm)
    status, opcoes = cmd.executar(servidor_acesso)

    validar_retorno_operacao_sem_lock(status, opcoes, ocultar_progresso)
    regioes = re.search(r'regioes: ([0-9]{1,})', opcoes).group(1)
    str_regioes = '\n'.join(string.ascii_uppercase[0:int(regioes)])
    opcoes = opcoes.replace('regioes: {}'.format(regioes), str_regioes)
    print('\n' + opcoes)


def imprimir_json_inventario(servidor_acesso, arquivo_inventario,
                             nome_vm, ocultar_progresso):
    configurar_vmm(servidor_acesso, ocultar_progresso)
    inventario_local = obter_inventario_local(
        servidor_acesso, arquivo_inventario, ocultar_progresso, filtro_nome_vm=nome_vm)

    adquirir_lock(servidor_acesso, inventario_local.agrupamento,
                  inventario_local.nuvem, ocultar_progresso)

    inventario_remoto = obter_inventario_remoto(
        servidor_acesso, inventario_local.agrupamento,
        inventario_local.nuvem, ocultar_progresso, filtro_nome_vm=nome_vm)

    liberar_lock(servidor_acesso, inventario_local.agrupamento,
                 inventario_local.nuvem, ocultar_progresso)

    imprimir_acao_corrente('Gerando JSON', ocultar_progresso)
    status, json_inventario = Inventario.get_json(
        inventario_local, inventario_remoto)
    validar_retorno_operacao_sem_lock(
        status, json_inventario, ocultar_progresso)
    print(json_inventario)


def planejar_sincronizacao(servidor_acesso, arquivo_inventario, ocultar_progresso):
    configurar_vmm(servidor_acesso, ocultar_progresso)
    inventario_local = obter_inventario_local(
        servidor_acesso, arquivo_inventario, ocultar_progresso)

    adquirir_lock(servidor_acesso,
                  inventario_local.agrupamento,
                  inventario_local.nuvem, ocultar_progresso)

    conteudo_arquivo = None
    plano_execucao = obter_plano_execucao(
        servidor_acesso, inventario_local, ocultar_progresso)

    if not plano_execucao.is_vazio():
        imprimir_acao_corrente('Gerando arquivo {}'.format(
            PlanoExecucao.ARQUIVO_PLANO_EXECUCAO), ocultar_progresso)
        status, conteudo_arquivo = plano_execucao.gerar_arquivo()
        validar_retorno_operacao_sem_lock(
            status, conteudo_arquivo, ocultar_progresso)

    liberar_lock(servidor_acesso, inventario_local.agrupamento,
                 inventario_local.nuvem, ocultar_progresso)

    if conteudo_arquivo:
        print('\nAlterações a serem realizadas:\n{}'.format(conteudo_arquivo))
    else:
        PlanoExecucao.excluir_arquivo()
        print(
            '\nNenhuma diferença encontrada entre o inventário local e o remoto.')


def executar_sincronizacao(servidor_acesso, arquivo_plano_execucao,
                           pular_confirmacao, arquivo_inventario,
                           ocultar_progresso):
    configurar_vmm(servidor_acesso, ocultar_progresso)

    # Obtendo plano de execução
    #
    # Caso de ter informado o plano de execução
    adquiriu_lock = False
    if arquivo_plano_execucao or os.path.isfile(PlanoExecucao.ARQUIVO_PLANO_EXECUCAO):
        imprimir_acao_corrente(
            'Carregando plano de execução', ocultar_progresso)
        status, plano_execucao = PlanoExecucao.carregar_plano_execucao(
            arquivo_plano_execucao or PlanoExecucao.ARQUIVO_PLANO_EXECUCAO)
        validar_retorno_operacao_sem_lock(
            status, plano_execucao, ocultar_progresso)
    # Caso de ter informado o arquivo de inventário
    elif arquivo_inventario:
        print('AVISO: plano de execução será calculado à partir do inventário.',
              flush=True)
        inventario_local = obter_inventario_local(
            servidor_acesso, arquivo_inventario, ocultar_progresso)

        adquirir_lock(servidor_acesso, inventario_local.agrupamento,
                      inventario_local.nuvem, ocultar_progresso)
        adquiriu_lock = True

        plano_execucao = obter_plano_execucao(
            servidor_acesso, inventario_local, ocultar_progresso)
    # Se não informou nem o plano nem o inventário, então informar erro
    else:
        finalizar_com_erro(
            'É necessário informar o plano de execução ou o arquivo de inventário.')

    if not adquiriu_lock:
        adquirir_lock(servidor_acesso, plano_execucao.agrupamento,
                      plano_execucao.nuvem, ocultar_progresso)

    if plano_execucao.is_vazio():
        liberar_lock(servidor_acesso, plano_execucao.agrupamento,
                     plano_execucao.nuvem, ocultar_progresso)
        print(
            '\nPlano de execução vazio: nada a alterar.')
    else:
        # Exibindo confirmação
        if not pular_confirmacao:
            print('\nAs seguintes operações serão executadas:')
            plano_execucao.imprimir_acoes()
            confirmar_acao_usuario_com_lock(
                servidor_acesso, plano_execucao.agrupamento, plano_execucao.nuvem)

        plano_execucao.executar(servidor_acesso, ocultar_progresso)

        liberar_lock(servidor_acesso, plano_execucao.agrupamento,
                     plano_execucao.nuvem, ocultar_progresso)
        plano_execucao.imprimir_resultado_execucao()


def remover_agrupamento_da_nuvem(servidor_acesso, agrupamento, nuvem,
                                 pular_confirmacao, ocultar_progresso):
    configurar_vmm(servidor_acesso, ocultar_progresso)

    adquirir_lock(servidor_acesso, agrupamento, nuvem, ocultar_progresso)

    inventario_remoto = obter_inventario_remoto(
        servidor_acesso, agrupamento, nuvem, ocultar_progresso)

    if not inventario_remoto.is_vazio():
        # Exibindo confirmação
        if not pular_confirmacao:
            print('\nAs seguintes máquinas serão excluídas:')
            for maquina_virtual in inventario_remoto.vms.values():
                print('ID_VMM: {}\tNome: {}\tStatus: {}'.format(
                    maquina_virtual.id_vmm,
                    maquina_virtual.nome,
                    maquina_virtual.status))
            confirmar_acao_usuario_com_lock(
                servidor_acesso, agrupamento, nuvem)

        plano_execucao = inventario_remoto.gerar_plano_exclusao()
        plano_execucao.executar(servidor_acesso, ocultar_progresso)

        liberar_lock(servidor_acesso, agrupamento, nuvem, ocultar_progresso)
        plano_execucao.imprimir_resultado_execucao()
    else:
        liberar_lock(servidor_acesso, agrupamento, nuvem, ocultar_progresso)
        print('\nNão há máquinas do agrupamento {} na nuvem {}: nada a fazer.'
              .format(agrupamento, nuvem))


def main():
    parser = get_parser()
    args = parser.parse_args()
    if not args.comando:
        finalizar_com_erro('Comando não informado. Use a opção -h para ajuda.')

    servidor_acesso = ServidorAcesso(
        args.servidor_acesso, args.usuario, args.senha, args.servidor)
    validar_conexao(servidor_acesso, args.ocultar_progresso)

    if args.comando == 'plan':
        planejar_sincronizacao(
            servidor_acesso, args.arquivo_inventario, args.ocultar_progresso)
    elif args.comando == 'apply':
        executar_sincronizacao(
            servidor_acesso, args.arquivo_plano_execucao,
            args.pular_confirmacao, args.arquivo_inventario,
            args.ocultar_progresso)
    elif args.comando == 'destroy':
        remover_agrupamento_da_nuvem(
            servidor_acesso, args.agrupamento, args.nuvem,
            args.pular_confirmacao, args.ocultar_progresso)
    elif args.comando == 'opts':
        listar_opcoes(servidor_acesso, args.ocultar_progresso)
    elif args.comando == 'show':
        imprimir_json_inventario(
            servidor_acesso, args.arquivo_inventario,
            args.nome_vm.upper(), args.ocultar_progresso)
